/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard;

import java.io.File;
import java.io.IOException;
import java.nio.ByteBuffer;

import org.simantics.databoard.accessor.Accessor;
import org.simantics.databoard.accessor.binary.BinaryObject;
import org.simantics.databoard.accessor.error.AccessorConstructionException;
import org.simantics.databoard.accessor.error.AccessorException;
import org.simantics.databoard.accessor.file.FileAccessor;
import org.simantics.databoard.accessor.file.FileLibrary;
import org.simantics.databoard.accessor.file.FileVariantAccessor;
import org.simantics.databoard.accessor.impl.AccessorParams;
import org.simantics.databoard.accessor.impl.DirectoryMap;
import org.simantics.databoard.accessor.java.JavaObject;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.error.BindingConstructionException;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.type.DataType;
import org.simantics.databoard.type.VariantType;
import org.simantics.databoard.util.binary.BinaryFile;
import org.simantics.databoard.util.binary.BinaryMemory;
import org.simantics.databoard.util.binary.RandomAccessBinary;

/**
 * This is a facade class for accessor services.
 *
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class Accessors {

	/**
	 * Open an accessor to byte data.
	 * 
	 * @param binary
	 * @param type
	 * @param params
	 * @return an accessor
	 * @throws AccessorConstructionException 
	 */
	@SuppressWarnings("unchecked")
	public static <T extends BinaryObject & Accessor> T getAccessor(RandomAccessBinary binary, DataType type, AccessorParams params) throws AccessorConstructionException {		
		return (T) BinaryObject.createAccessor(binary, type, params);
	}
	
	/**
	 * Open an accessor to byte data.
	 * 
	 * @param binary
	 * @param type
	 * @return an accessor
	 * @throws AccessorConstructionException 
	 */
	@SuppressWarnings("unchecked")
	public static <T> T getAccessor(RandomAccessBinary binary, DataType type) throws AccessorConstructionException {		
		return (T) BinaryObject.createAccessor(binary, type, AccessorParams.DEFAULT);
	}
	
	/**
	 * Open an accessor to byte data.
	 * 
	 * @param binary
	 * @param type
	 * @return an accessor
	 * @throws AccessorConstructionException 
	 */
	@SuppressWarnings("unchecked")
	public static <T extends BinaryObject & Accessor> T getAccessor(byte[] binary, DataType type) throws AccessorConstructionException {
		RandomAccessBinary rab = new BinaryMemory( ByteBuffer.wrap(binary) );
		return (T) BinaryObject.createAccessor(rab, type, AccessorParams.DEFAULT);
	}

	/**
	 * Open an accessor to a Java Object.
	 * <p>
	 * Accessor is disposed by leaving it to garbage collector.
	 * <p>
	 * Do not modify the object outside the accessor as long as you use the accessor.
	 * Exterioir modifications will mix up listening the mechanism.
	 * Also, do not create more than one accessor to one Object. 
	 * <p>
	 * The accessor is not multi-thread accessible. You must provide mutual exclusion
	 * locking mechanism to the whole Object Model, if you intend to use it in 
	 * multi-thread environment. 
	 * 
	 * @since 0.5
	 * @param binding
	 * @param value
	 * @return accessor
	 * @throws AccessorConstructionException
	 */
	@SuppressWarnings("unchecked")
	public static <T extends JavaObject & Accessor> T getAccessor(Binding binding, Object value, AccessorParams params) 
	throws AccessorConstructionException {
		return (T) JavaObject.createAccessor(null, binding, value, params);
	}
	
	/**
	 * Open an accessor to a Java Object.
	 * <p>
	 * Accessor is disposed by leaving it to garbage collector.
	 * <p>
	 * Do not modify the object outside the accessor as long as you use the accessor.
	 * Exterioir modifications will mix up listening the mechanism.
	 * Also, do not create more than one accessor to one Object. 
	 * <p>
	 * The accessor is not multi-thread accessible. You must provide mutual exclusion
	 * locking mechanism to the whole Object Model, if you intend to use it in 
	 * multi-thread environment. 
	 * 
	 * @since 0.5
	 * @param binding
	 * @param value
	 * @return accessor
	 * @throws AccessorConstructionException
	 */
	@SuppressWarnings("unchecked")
	public static <T extends JavaObject & Accessor> T getAccessor(Binding binding, Object value) 
	throws AccessorConstructionException {
		return (T) JavaObject.createAccessor(null, binding, value, AccessorParams.DEFAULT);
	}
	
	/**
	 * Open Accessor to a Java Object. This version reads the type using reflection.
	 * <p>
	 * Accessor is disposed by leaving it to garbage collector.
	 * <p>
	 * Do not modify the object outside the accessor as long as you use the accessor.
	 * Exterioir modifications will mix up listening the mechanism.
	 * Also, do not create more than one accessor to one Object. 
	 * <p>
	 * The accessor is not multi-thread accessible. You must provide mutual exclusion
	 * locking mechanism to the whole Object Model, if you intend to use it in 
	 * multi-thread environment. 
	 * 
	 * @since 0.5
	 * @param value
	 * @return accessor
	 * @throws AccessorConstructionException
	 */
	@SuppressWarnings("unchecked")
	public static <T extends JavaObject & Accessor> T getAccessor(Object value) 
	throws AccessorConstructionException {
		try {
			Binding binding = Bindings.getBinding(value.getClass());
			return (T) JavaObject.createAccessor(null, binding, value, AccessorParams.DEFAULT);
		} catch (BindingConstructionException e) {
			throw new AccessorConstructionException(e);
		}
	}
	
	/**
	 * Open an accessor to a binary file (.dbb). The file is always a variant.
	 * The accessor must be closed by invoking {@link FileAccessor#close()} at 
	 * root or any sub-accessor.
	 * <p>
	 * To share accessors of the same file use {@link FileLibrary} utility.
	 * 
	 * @since 0.5
	 * @param file
	 * @return file accessor
	 * @throws AccessorConstructionException 
	 */
	public static FileVariantAccessor openAccessor(File file) throws AccessorConstructionException {		
		try {			
			BinaryFile bf = new BinaryFile(file);
			FileVariantAccessor result = (FileVariantAccessor) BinaryObject.createAccessor(bf, VariantType.INSTANCE, AccessorParams.DEFAULT);
			return result;
		} catch (IOException e) {
			throw new AccessorConstructionException(e);
		}
	}
	
	/**
	 * Create a new binary file with an empty value and open an accessor.
	 * If file already exists, it is overwritten.
	 * It is recommended that the file extension is .dbb (Databoard Binary)
	 * <p>
	 * The caller must close the file with {@link FileAccessor#close()}.  
	 * 
	 * @since 0.5
	 * @param file
	 * @return an accessor to the root variant node
	 */
	public static FileVariantAccessor createFile(File file) 
	throws AccessorConstructionException {
		try {
			file.createNewFile();
			BinaryFile bf = new BinaryFile(file);
			FileVariantAccessor result = (FileVariantAccessor) BinaryObject.createAccessor(bf, VariantType.INSTANCE, AccessorParams.DEFAULT);
			
			// Write initial value
			try {
				Binding vb = Bindings.getBinding(void.class);
				Object vv = vb.createDefault();
				result.setContentValue(vb, vv);
			} catch (AccessorException e) {
				// Close file
				try { result.close(); } catch (AccessorException e1) {}
				// Rethrow exception
				throw new AccessorConstructionException(e);
			} catch (BindingConstructionException e) {
				// Close file
				try { result.close(); } catch (AccessorException e1) {}
				// Rethrow exception
				throw new AccessorConstructionException(e);
			} catch (BindingException e) {
				// Close file
				try { result.close(); } catch (AccessorException e1) {}
				// Rethrow exception
				throw new AccessorConstructionException(e);
			} 
			
			return result;
		} catch (IOException e) {
			throw new AccessorConstructionException(e);
		}
	}	

	/**
	 * Create a binary file (.dbb) and open an accessor. The file is always a variant.
	 * <p>
	 * File is closed with {@link FileAccessor#close()} at root or any 
	 * sub-accessor.
	 * <p>
	 * To share accessors of the same file use {@link FileLibrary} utility.
	 * 
	 * @param <T>
	 * @param file
	 * @param type
	 * @return an accessor to the value node
	 * @throws AccessorConstructionException 
	 */
	@SuppressWarnings("unchecked")
	public static <T extends FileAccessor> T createFile(File file, DataType type) throws AccessorConstructionException
	{
		boolean existed;
		BinaryFile bf;
		try {
			existed = file.createNewFile();
			bf = new BinaryFile(file);
		} catch (IOException e1) {
			throw new AccessorConstructionException(e1);
		}
		
		try {
			FileVariantAccessor result = (FileVariantAccessor) BinaryObject.createAccessor(bf, VariantType.INSTANCE, AccessorParams.DEFAULT);
			Binding binding = Bindings.getBinding(type);
			Object value = binding.createDefault();
			result.setContentValue(binding, value);
			return (T) result.getContentAccessor();
		} catch (BindingException e) {
			if (!existed) file.delete();
			throw new AccessorConstructionException(e);
		} catch (AccessorException e) {
			if (!existed) file.delete();
			throw new AccessorConstructionException(e);
		}
	}
	
	/**
	 * Open a Map(Variant, Variant) accessor to a directory.
	 * Map entries are binary files (.dbb), and the file name is the key with the following encoding: 
	 *   
	 *  Filenames have the following encoding:
	 *    S<string>.dbb   String types
	 *                        Control characters " : < > | ? * \ / % [0..31] [128..] are escaped as %<hex><hex>
	 *                        " "-space is _
	 *    I<integer>.dbb  Integer types
	 *    L<long>.dbb     Long types
	 *    H<hex>.dbb      All other cases the value as binary 
	 * 
	 * The caller must close (FolderMap#close()) the accessor after usage. 
	 * This releases file handles and a directory polling thread.
	 * 
	 * FolderMap is not concurrent-use-safe. Its files are not either.
	 * The user must ensure that the files are not used concurrently from 
	 * different threads.
	 * 
	 * @param directory
	 * @return MapAccessor
	 */
	public static DirectoryMap openDirectory(File directory) {
		return new DirectoryMap(directory);
	}	
		
}

