/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.bindingscheme;

import java.util.Map;
import java.util.WeakHashMap;

import org.simantics.databoard.Bindings;
import org.simantics.databoard.binding.ArrayBinding;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.BooleanBinding;
import org.simantics.databoard.binding.ByteBinding;
import org.simantics.databoard.binding.DoubleBinding;
import org.simantics.databoard.binding.FloatBinding;
import org.simantics.databoard.binding.IntegerBinding;
import org.simantics.databoard.binding.LongBinding;
import org.simantics.databoard.binding.MapBinding;
import org.simantics.databoard.binding.OptionalBinding;
import org.simantics.databoard.binding.RecordBinding;
import org.simantics.databoard.binding.StringBinding;
import org.simantics.databoard.binding.UnionBinding;
import org.simantics.databoard.binding.VariantBinding;
import org.simantics.databoard.binding.Binding.Visitor;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.error.RuntimeBindingException;
import org.simantics.databoard.type.ArrayType;
import org.simantics.databoard.type.ByteType;
import org.simantics.databoard.type.DoubleType;
import org.simantics.databoard.type.FloatType;
import org.simantics.databoard.type.IntegerType;
import org.simantics.databoard.type.LongType;
import org.simantics.databoard.type.StringType;
import org.simantics.databoard.util.Range;

/**
 * Visitor that creates a default instance of a DataType.
 * This visitor may throw RuntimeBindingException. 
 * 
 * Type                     Value
 * ------------------------------------------------------
 * Boolean                  false
 * Byte, Integer, Long      0 or min limit
 * Float, Double            0.0 or min limit
 * String                   ""
 * Optional                 *novalue*
 * Union                    tag 0
 * Record                   each field with default value
 * Array                    min range number of elements
 * Map                      no entries
 * Variant                  {} : void
 * 
 * 
 * TODO Create String according to the pattern
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class DefaultValue implements Visitor<Object> {

	/** Map of default values already created. Used to link back to recursive records */
	Map<Binding, Object> map = new WeakHashMap<Binding, Object>(1); 
		
	@Override
	public Object visit(ArrayBinding b) {
		Object result = map.get(b);
		if (result!=null) return result;
		
		ArrayType at = b.type();
		Range r = at.getLength();
		int min = (r!=null) ? r.getLower().getValue().intValue() : 0;
			
		if (min>0) {
			Binding componentBinding = b.getComponentBinding();
			Object[] array = new Object[min];
			for (int i=0; i<min; i++) {
				array[i] = componentBinding.accept(this);
			}
			result = b.createUnchecked(array);
		} else {
			result = b.createUnchecked();
		}		
		map.put(b, result);
		return result;
	}

	@Override
	public Object visit(BooleanBinding b) {
		Object result = map.get(b);
		if (result!=null) return result;
		result = b.createUnchecked(Boolean.FALSE);
		map.put(b, result);
		return result;
	}

	@Override
	public Object visit(DoubleBinding b) {
		Object result = map.get(b);
		if (result!=null) return result;
		DoubleType type = b.type();
		Range range = type.getRange();
		double value = (range==null) ? 0.0 : range.getLower().getValue().doubleValue();
		result = b.createUnchecked(value);
		map.put(b, result);
		return result;
	}

	@Override
	public Object visit(FloatBinding b) {
		Object result = map.get(b);
		if (result!=null) return result;
		FloatType type = b.type();
		Range range = type.getRange();
		float value = (range==null) ? 0.f : range.getLower().getValue().floatValue();
		result = b.createUnchecked(value);
		map.put(b, result);
		return result;
	}

	@Override
	public Object visit(IntegerBinding b) {
		Object result = map.get(b);
		if (result!=null) return result;
		IntegerType type = b.type();
		Range range = type.getRange();
		int value = (range==null) ? 0 : range.getLower().getValue().intValue();
		result = b.createUnchecked(value);
		map.put(b, result);
		return result;
	}

	@Override
	public Object visit(ByteBinding b) {
		Object result = map.get(b);
		if (result!=null) return result;
		ByteType type = b.type();
		Range range = type.getRange();
		byte value = (range==null) ? 0 : range.getLower().getValue().byteValue();
		result = b.createUnchecked(value);
		map.put(b, result);
		return result;
	}

	@Override
	public Object visit(LongBinding b) {
		Object result = map.get(b);
		if (result!=null) return result;
		LongType type = b.type();
		Range range = type.getRange();
		long value = (range==null) ? 0 : range.getLower().getValue().longValue();
		result = b.createUnchecked(value);
		map.put(b, result);
		return result;
	}

	@Override
	public Object visit(OptionalBinding b) {
		Object result = map.get(b);
		if (result!=null) return result;
		result = b.createNoValueUnchecked();
		map.put(b, result);
		return result;
	}

	@Override
	public Object visit(RecordBinding b) {
		try {
			Object result = map.get(b);
			if (result!=null) return result;
			Object[] values = new Object[ b.count() ];
			result = b.createPartial();
			map.put(b, result);
			for (int i=0; i<values.length; i++) {
				Binding cb = b.getComponentBinding(i);			
				values[i] = cb.accept(this);
			}
			b.setComponents(result, values);
			return result;
		} catch (BindingException e) {
			throw new RuntimeBindingException(e);
		}
	}

	@Override
	public Object visit(StringBinding b) {
		Object result = map.get(b);
		if (result!=null) return result;
		StringType st = b.type();
		if (st.getPattern() != null) {
			// TODO Create a string that is valid to the pattern
			result = b.createUnchecked("");			
		} else {
			result = b.createUnchecked("");
		}
		map.put(b, result);
		return result;
	}

	@Override
	public Object visit(UnionBinding b) {
		Object result = map.get(b);
		if (result!=null) return result;
		int tag = 0;
		Binding componentBinding = b.getComponentBinding(tag);
		Object tag0defaultValue = componentBinding.accept(this);
		result = b.createUnchecked(tag, tag0defaultValue);
		map.put(b, result);
		return result;
	}

	
	@Override
	public Object visit(VariantBinding b) {
		Object result = map.get(b);
		if (result!=null) return result;
		Binding voidBinding = Bindings.getBindingUnchecked(void.class);
		Object voidValue = null;
		result = b.createUnchecked(voidBinding, voidValue);
		map.put(b, result);
		return result;
	}

	@Override
	public Object visit(MapBinding b) {
		Object result = map.get(b);
		if (result!=null) return result;
		result = b.createUnchecked();
		map.put(b, result);
		return result;
	}
    
}

