/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.serialization;

import gnu.trove.TObjectIntHashMap;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.List;

import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.util.binary.BinaryFile;
import org.simantics.databoard.util.binary.BinaryReadable;
import org.simantics.databoard.util.binary.BinaryWriteable;
import org.simantics.databoard.util.binary.ByteBufferReadable;
import org.simantics.databoard.util.binary.ByteBufferWriteable;
import org.simantics.databoard.util.binary.InputStreamReadable;
import org.simantics.databoard.util.binary.OutputStreamWriteable;


public abstract class Serializer {

	/**
	 * Finalize 2-phased construction. 
	 * 
	 * @throws RuntimeException thrown if construction fails
	 */
	public void postConstruction(SerializationFormat format)
	throws SerializerConstructionException {		
	}
	
	/**
	 * Serialize obj to out.
	 * 
	 * The identities argument is a map of identities (in the binary block) and objects
	 * that have already been serialized. Once serialized, an object is added to the map.
	 * Typically an empty map is provided. If the type has no recursion, a <code>null</code>
	 * value can be provided.  
	 * 
	 * @param out
	 * @param identities Thread local empty map or <code>null</code> if there is no recursion
	 * @param obj
	 * @throws IEncodingException
	 * @throws IOException
	 * @throws BindingException
	 */
	public abstract void serialize(BinaryWriteable out, TObjectIntHashMap<Object> identities, Object obj) throws SerializationException, IOException, BindingException;
	public abstract void serialize(BinaryWriteable out, Object obj) throws SerializationException, IOException, BindingException;
	
	/**
	 * Deserialize an object from a readable.
	 * 
	 * The identities argument is a list of identities (in the binary block) of objects
	 * that have already been deserialized. Once deserialized they are added to the list.
	 * Typically an empty list is provided. If the type has no recursion, a <code>null</code>
	 * value can be provided.  
	 * 
	 * @param in 
	 * @param identities empty identities array or <code>null</code> if there is no recursion
	 * @return the instance
	 * @throws SerializationException
	 * @throws IOException
	 * @throws BindingException Could not bind the data into an object 
	 */
	public abstract Object deserialize(BinaryReadable in, List<Object> identities) throws SerializationException, IOException, BindingException;
	public abstract Object deserialize(BinaryReadable in) throws SerializationException, IOException, BindingException;

	
	/**
	 * Deserialize the next object in an input stream.
	 * Note, if multiple objects are deserialized from the same stream, it is 
	 * more efficient to instantiate InputStreamReadable and identities only once.
	 * 
	 * @param in
	 * @return The object deserialized into a Java Object
	 * @throws SerializationException
	 * @throws IOException
	 * @throws BindingException
	 */
	public Object deserialize(InputStream in) throws SerializationException, IOException, BindingException
	{
		InputStreamReadable read = new InputStreamReadable(in, Long.MAX_VALUE);
		read.order( ByteOrder.LITTLE_ENDIAN );
		return deserialize(read);
	}
	
	/**
	 * Deserialize object from a file
	 * 
	 * @param file 
	 * @return the object
	 * @throws SerializationException
	 * @throws IOException
	 * @throws BindingException
	 */
	public Object deserialize(File file) throws SerializationException, IOException, BindingException
	{
		BinaryFile f = new BinaryFile(file);
		try {
			f.order( ByteOrder.LITTLE_ENDIAN );
			return deserialize(f);
		} finally {
			f.close();
		}		
	}
	
	/**
	 * Deserialize an object in byte[] format.
	 * 
	 * @param data
	 * @return the instance
	 * @throws SerializationException
	 * @throws IOException
	 * @throws BindingException
	 */
	public Object deserialize(byte[] data) 
	throws SerializationException, IOException, BindingException
	{
		ByteBuffer buffer = ByteBuffer.wrap( data );
		BinaryReadable readable = new ByteBufferReadable( buffer );
		readable.order( ByteOrder.LITTLE_ENDIAN );
		return deserialize(readable);
	}

	/**
	 * Skip over an object in a stream. This method deserializes the object 
	 * without producing a result or reading thru all bytes. 
	 * 
	 * @param in
	 * @param identities
	 * @throws SerializationException
	 * @throws IOException
	 * @throws BindingException
	 */
	public abstract void skip(BinaryReadable in, List<Object> identities) throws SerializationException, IOException;
	public abstract void skip(BinaryReadable in) throws SerializationException, IOException;
	
	/**
	 * Skip over an object in a stream. This method deserializes the object 
	 * without producing a result or reading thru all bytes. 
	 * 
	 * @param in
	 * @throws SerializationException
	 * @throws IOException
	 * @throws BindingException
	 */	
	public void skip(InputStream in) throws SerializationException, IOException, BindingException
	{
		InputStreamReadable read = new InputStreamReadable(in, Long.MAX_VALUE);
		read.order( ByteOrder.LITTLE_ENDIAN );
		skip(read);
	}
	
	/**
	 * Get constant size of the data type in its binary serialized format 
	 * 
	 * @return size in bytes or null if not fixed
	 */
	public abstract Integer getConstantSize();
	
	/**
	 * 
	 * @param obj
	 * @param identities thread local empty hash map
	 * @return number of bytes required to serialize obj
	 * @throws SerializationException 
	 * @throws BindingException 
	 */
	public abstract int getSize(Object obj, TObjectIntHashMap<Object> identities) throws SerializationException, BindingException;
	public abstract int getSize(Object obj) throws SerializationException, BindingException;
	
	/**
	 * Serializes an object to a byte[].
	 * 
	 * @param obj
	 * @return byte array containing the obj in its serialized format. 
	 * @throws SerializationException 
	 * @throws IOException
	 * @throws BindingException
	 */
	public byte[] serialize(Object obj)
	throws SerializationException, BindingException
	{
		try {
			TObjectIntHashMap<Object> identities = new TObjectIntHashMap<Object>();
			int size = getSize(obj, identities);
			identities.clear();
			ByteBuffer buffer = ByteBuffer.allocate( size );
			buffer.order( ByteOrder.LITTLE_ENDIAN );
			BinaryWriteable writable = new ByteBufferWriteable( buffer );
			serialize(writable, identities, obj);
			buffer.rewind();
			return buffer.array();
		} catch (IOException e) {
			// Unexpected
			throw new RuntimeException("Unexpected IOException", e);
		}
	}
	
	/**
	 * Serializes an object to an output stream.
	 * Note, if multiple objects are serialized to the same stream, it is 
	 * more efficient to instantiate OutputStreamWriteable and identities only once.
	 * 
	 * @param obj
	 * @param out
	 * @throws SerializationException
	 * @throws IOException
	 * @throws BindingException
	 */
	public void serialize(Object obj, OutputStream out) throws SerializationException, IOException, BindingException
	{
		OutputStreamWriteable writ = new OutputStreamWriteable(out);
		writ.order( ByteOrder.LITTLE_ENDIAN );
		TObjectIntHashMap<Object> identities = new TObjectIntHashMap<Object>();
		serialize(writ, identities, obj);
	}
	
	/**
	 * Serialize an object to a file 
	 * 
	 * @param obj
	 * @param file
	 * @throws SerializationException 
	 * @throws IOException
	 * @throws BindingException
	 */
	public void serialize(Object obj, File file)
	throws SerializationException, BindingException
	{
		try {
			TObjectIntHashMap<Object> identities = new TObjectIntHashMap<Object>();
			int size = getSize(obj, identities);
			identities.clear();
			BinaryFile writable = new BinaryFile( file );
			try {
				writable.order( ByteOrder.LITTLE_ENDIAN );			
				serialize(writable, identities, obj);
			} finally {
				writable.close();
			}
		} catch (IOException e) {
			// Unexpected
			throw new RuntimeException("Unexpected IOException", e);
		}
	}	

	/**
	 * Get object as readable Input Stream.
	 * 
	 * @param obj
	 * @return input stream
	 * @throws SerializationException
	 * @throws BindingException
	 */
	public InputStream getInputStream(Object obj) throws SerializationException, BindingException
	{
		// Trivial implementation
		byte[] data = serialize(obj);
		return new ByteArrayInputStream(data);
	}
	
	/**
	 * Serializer for recursive data types
	 */
	public static abstract class RecursiveSerializer extends Serializer {
		public void serialize(BinaryWriteable out, Object obj) throws SerializationException, IOException, BindingException {
			TObjectIntHashMap<Object> identities = new TObjectIntHashMap<Object>(0);
			serialize(out, identities, obj);
		}
		public Object deserialize(BinaryReadable in) throws SerializationException, IOException, BindingException {
			List<Object> identities = new ArrayList<Object>(0);
			return deserialize(in, identities);
		}
		@Override
		public void skip(BinaryReadable in) throws SerializationException, IOException {
			List<Object> identities = new ArrayList<Object>(0);
			skip(in, identities);
		}
		@Override
		public int getSize(Object obj)
				throws SerializationException, BindingException {
			TObjectIntHashMap<Object> identities = new TObjectIntHashMap<Object>(0);
			return getSize(obj, identities);
		}
	}
	
	/**
	 * Serializer for non-recursive data types
	 */
	public static abstract class NonRecursiveSerializer extends Serializer {
		public void serialize(BinaryWriteable out, TObjectIntHashMap<Object> identities, Object obj) throws SerializationException, IOException, BindingException {
			serialize(out, obj);
		}
		public Object deserialize(BinaryReadable in, List<Object> identities) throws SerializationException, IOException, BindingException {
			return deserialize(in);
		}
		@Override
		public void skip(BinaryReadable in, List<Object> identities) throws SerializationException, IOException {
			skip(in);
		}
		@Override
		public int getSize(Object obj, TObjectIntHashMap<Object> identities) throws SerializationException, BindingException {
			return getSize(obj);
		}
	}
	
	/**
	 * Serializer for composite data types
	 */
	public static abstract class CompositeSerializer extends Serializer {
		boolean recursive;
		protected CompositeSerializer(boolean recursive) {
			this.recursive = recursive;
		}
		public void serialize(BinaryWriteable out, Object obj) throws SerializationException, IOException, BindingException {
			TObjectIntHashMap<Object> identities = recursive ? new TObjectIntHashMap<Object>(0) : null;
			serialize(out, identities, obj);
		}
		public Object deserialize(BinaryReadable in) throws SerializationException, IOException, BindingException {
			List<Object> identities = recursive ? new ArrayList<Object>(0) : null;
			return deserialize(in, identities);
		}
		@Override
		public void skip(BinaryReadable in) throws SerializationException, IOException {
			List<Object> identities = recursive ? new ArrayList<Object>(0) : null;
			skip(in, identities);
		}
		@Override
		public int getSize(Object obj)
				throws SerializationException, BindingException {
			TObjectIntHashMap<Object> identities = recursive ? new TObjectIntHashMap<Object>(0) : null;
			return getSize(obj, identities);
		}
	}
	
}

