/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.type;

import java.util.HashSet;
import java.util.Set;

import org.simantics.databoard.annotations.Referable;
import org.simantics.databoard.util.IdentityPair;

public @Referable class UnionType extends DataType {
	
    public Component[] components;
    
    public UnionType() {}
    
    public UnionType(Component...components) {
    	if (components.length==0) throw new IllegalArgumentException("need atleast 1 tag type");
    	this.components = components;
    }
    
    public void addComponent(String name, DataType type)
    {
    	Component c = new Component(name, type);
    	if (components == null) {
    		components = new Component[] { c };
    	} else {
    		Component[] newComponents = new Component[ components.length +1 ];
    		System.arraycopy(components, 0, newComponents, 0, components.length);
    		newComponents[ components.length ] = c;
    		components = newComponents;    	
    	}    	
    }
    
    public void removeComponent(String tagName) {
    	int index = getTagIndex2(tagName);
    	if (index<0) return;
		Component[] newComponents = new Component[ components.length -1 ];
		if (index>0) System.arraycopy(components, 0, newComponents, 0, index);
		if (index<newComponents.length) System.arraycopy(components, index+1, newComponents, index, newComponents.length - index);
		components = newComponents;    	
		// xxx untested
    }
    
    /**
     * Get tag by name.
     * 
     * @param fieldName component name
     * @return component index or <code>null</code> if one does not exist
     */
    public Integer getTagIndex(String fieldName) {
        for (int i=0; i<components.length; i++)
            if (components[i].name.equals(fieldName)) return i;
        return null;
    }
    
    /**
     * Get tag by name.
     * 
     * @param fieldName component name
     * @return component index or -1 if one does not exist
     */
    public int getTagIndex2(String fieldName) {
        for (int i=0; i<components.length; i++)
            if (components[i].name.equals(fieldName)) return i;
        return -1;
    }
    

    /**
     * Get component Datatype by field name
     * @param fieldName
     * @return datatype or <code>null</code>
     */
    public DataType getComponentType(String fieldName) {
    	int index = getTagIndex2(fieldName);
    	if (index<0) return null;
    	return components[index].type;
    }
    
    /**
     * Get component by name.
     * 
     * @param fieldName component name
     * @return component or <code>null</code> if one does not exist
     */
    public Component getComponent(String fieldName) {
        for (Component c : components)
            if (c.name.equals(fieldName)) return c;
        return null;
    }
    
    @Override
    protected void collectSubtypes(Set<DataType> subtypes,
    		Set<DataType> recursiveSubtypes) {
    	if(!subtypes.add(this)) {
    		recursiveSubtypes.add(this);
    		return;
    	}
    	for(Component component : components)
	        component.type.collectSubtypes(subtypes, recursiveSubtypes);
    }
    
    @Override
    protected boolean deepEquals(Object obj,
    		Set<IdentityPair<DataType, DataType>> compareHistory) {    	
		if (this==obj) return true;
		if (obj instanceof UnionType == false) return false;
		UnionType other = (UnionType) obj;
				
		if (components.length!= other.components.length) return false;
		// Verify names
		for (int i = 0; i<components.length; i++) {
			Component lc = components[i];
			Component rc = other.components[i];
			if (!lc.name.equals(rc.name)) return false;
			
		}

		// Verify types
		if (compareHistory==null) compareHistory = new HashSet<IdentityPair<DataType, DataType>>(1);

		IdentityPair<DataType, DataType> pair = new IdentityPair<DataType, DataType>(this, other);
		if (compareHistory.contains(pair)) return true;
		compareHistory.add(pair);
		
		for (int i = 0; i<components.length; i++) {
			Component lc = components[i];
			Component rc = other.components[i];
			if (!lc.type.deepEquals(rc.type, compareHistory)) return false;
		}
		return true;
	}
	
	@Override
	public int hashCode() {
		int hash = 42342345;
		for (Component c : components) 
			hash = hash*7 + c.name.hashCode();
		return hash;
	}
    
	@Override
	public void accept(Visitor1 v, Object obj) {
	    v.visit(this, obj);        
	}

	@Override
	public <T> T accept(Visitor<T> v) {
	    return v.visit(this);
	}
	
	public int count() {
		return components.length;
	}
	
	public Integer getComponentIndex(String tagName) {
		for (int i=0; i<components.length; i++) {
			if (components[i].name.equals(tagName))
				return i;
		}
		return null;
	}
	
	public Component getComponent(int i) {
		return components[i];
	}
	
	public Component[] getComponents() {
		return components;
	}
	
	
}
