/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.util.binary;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

import org.simantics.databoard.file.RandomAccessBinaryList;

/**
 * Random access binary is a interface to access a binary object (BLOB).
 * Binary object is random accessible and has a length.
 * It can be read and written, enlarged and shrunk at any position.
 * <p>
 * There is a pointer which moves on every read and write operation. It can 
 * moved with {@link #position()}. If bytes are written and the pointer is 
 * moved outside the blob or moves while writing, the blob is enlarged.
 * 
 * @see Blob Cascading implementation
 * @see BinaryFile RandomAccessFile implementation
 * @see RandomAccessBinaryList java.util.List implementation 
 * @author Toni Kalajainen
 */
public interface RandomAccessBinary extends BinaryWriteable, BinaryReadable {

	/**
	 * Write a byte
	 * 
	 * @param b
	 * @throws IOException
	 */
    void put(byte b)
    throws IOException;
    
    /**
     * Put n bytes from the remaining of the byte array. 
     * This operation moves the pointer in byte buffer.
     * 
     * @param src
     * @throws IOException
     */
    void put(ByteBuffer src)
    throws IOException;
    
    /**
     * Put n bytes from the remaining of the byte buffer. 
     * This operation moves the pointer in byte buffer.
     * 
     * @param src
     * @param length
     * @throws IOException
     */
    void put(ByteBuffer src, int length)
    throws IOException;
    
    void put(byte[] src, int offset, int length)
    throws IOException;
    
    void put(byte[] src)
    throws IOException;
    
    void putShort(short value)
    throws IOException;
    
    void putInt(int value)
    throws IOException, IndexOutOfBoundsException;
    
    void putLong(long value)
    throws IOException;
    
    void putFloat(float value)
    throws IOException;
    
    void putDouble(double value)
    throws IOException;

    /**
     * Get byte order
     * 
     * @return byte order
     */
	ByteOrder order();

	/**
	 * Set a new byte order
	 * 
	 * @param order new byte order
	 */
	void order(ByteOrder order);

	/**
	 * Flush write buffer 
	 * 
	 * @throws IOException
	 */
	void flush()
	throws IOException;

    byte get()
    throws IOException;

    void get(byte[] dst, int offset, int length)
    throws IOException;
    
    void get(byte[] dst)
    throws IOException;
    
    /**
     * Get buf fully 
     * 
     * @param buf
     * @throws IOException
     */
    void get(ByteBuffer buf)
    throws IOException;
    
    /**
     * Get fully length bytes
     * 
     * @param buf
     * @param length
     * @throws IOException
     */
    void get(ByteBuffer buf, int length)
    throws IOException;
    
    short getShort()
    throws IOException;
    
    int getInt()
    throws IOException;
    
    long getLong()
    throws IOException;
    
    float getFloat()
    throws IOException;
    
    double getDouble()
    throws IOException;
    	
	/**
	 * Remove bytes from the position of the current pointer. If the pointer is 
	 * outside the blob, an error is thrown. 
	 * 
	 * @param bytes
	 * @throws IOException
	 */
	void removeBytes(long bytes) throws IOException;
	
	/**
	 * Insert bytes at current pointer. If the pointer is outside the size of
	 * the blob, the blob is enlarged to fit the insertion. 
	 * 
	 * The operation does not move pointer.
	 * 
	 * @param bytes
	 * @throws IOException
	 */
	void insertBytes(long bytes) throws IOException;
	
	/**
	 * Set length of the binary object.
	 *   
	 * @param newLength
	 * @throws IOException
	 */
	void setLength(long newLength) throws IOException;
	
	/**
	 * Get the length of the binary object.
	 * 
	 * @return the length
	 * @throws IOException
	 */
	long length() throws IOException;
	
	/**
	 * Flush and close the blob. 
	 */
	void close() throws IOException;

	/**
	 * Set new cursor position.
	 * 
	 * The position can be set outside the binary object.
	 * 
	 * @param newPosition 
	 */
	void position(long newPosition) throws IOException;

	/**
	 * Get the position of the cursor
	 * 
	 * @return cursor 
	 */
	long position() throws IOException;

	/**
	 * Move pointer forward
	 */
	void skip(long bytes) throws IOException;	
	
}

